/* mRss - Copyright (C) 2005 bakunin - Andrea Marchesini 
 *                                <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "mrss.h"
#include "mrss_internal.h"

static size_t
__mrss_memorize_file (void *ptr, size_t size, size_t nmemb, void *data)
{
  register int realsize = size * nmemb;
  __mrss_download_t *mem = (__mrss_download_t *) data;

  if (!mem->mm)
    {
      if (!(mem->mm = (char *) malloc (realsize + 1)))
	return -1;
    }
  else
    {
      if (!(mem->mm = (char *) realloc (mem->mm, mem->size + realsize + 1)))
	return -1;
    }

  memcpy (&(mem->mm[mem->size]), ptr, realsize);
  mem->size += realsize;
  mem->mm[mem->size] = 0;

  return realsize;
}

__mrss_download_t *
__mrss_download_file (char *fl)
{
  __mrss_download_t *chunk;
  CURL *curl;

  if (!(chunk = (__mrss_download_t *) malloc (sizeof (__mrss_download_t))))
    return NULL;

  chunk->mm = NULL;
  chunk->size = 0;

  curl_global_init (CURL_GLOBAL_DEFAULT);
  if (!(curl = curl_easy_init ()))
    {
      free (chunk);
      return NULL;
    }

  curl_easy_setopt (curl, CURLOPT_URL, fl);
  curl_easy_setopt (curl, CURLOPT_WRITEFUNCTION, __mrss_memorize_file);
  curl_easy_setopt (curl, CURLOPT_FILE, (void *) chunk);
  curl_easy_setopt (curl, CURLOPT_TIMEOUT, 10);

  if (curl_easy_perform (curl))
    {
      if (chunk->mm)
	free (chunk->mm);

      free (chunk);

      curl_easy_cleanup (curl);
      curl_global_cleanup ();

      return NULL;
    }

  curl_easy_cleanup (curl);
  curl_global_cleanup ();

  return chunk;
}

/* EOF */
