/* mRss - Copyright (C) 2005 bakunin - Andrea Marchesini 
 *                                <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "mrss.h"
#include "mrss_internal.h"

static void
__mrss_parser_rss_image (xmlDocPtr doc, xmlNodePtr cur, mrss_t * data)
{
  char *c;

  for (cur = cur->children; cur; cur = cur->next)
    {
      if (cur->type == XML_ELEMENT_NODE)
	{
	  /* title */
	  if (!xmlStrcmp (cur->name, (xmlChar *) "title")
	      && !data->image_title
	      && (c =
		  (char *) xmlNodeListGetString (doc, cur->xmlChildrenNode,
						 1)))
	    data->image_title = c;

	  /* url */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "url")
		   && !data->image_url
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->image_url = c;

	  /* link */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "link")
		   && !data->image_link
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->image_link = c;

	  /* width */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "width")
		   && !data->image_width
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->image_width = atoi (c);

	  /* height */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "height")
		   && !data->image_height
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->image_height = atoi (c);

	  /* description */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "description")
		   && !data->image_description
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->image_description = c;
	}
    }
}

static void
__mrss_parser_rss_textinput (xmlDocPtr doc, xmlNodePtr cur, mrss_t * data)
{
  char *c;

  for (cur = cur->children; cur; cur = cur->next)
    {
      if (cur->type == XML_ELEMENT_NODE)
	{
	  /* title */
	  if (!xmlStrcmp (cur->name, (xmlChar *) "title")
	      && !data->textinput_title
	      && (c =
		  (char *) xmlNodeListGetString (doc, cur->xmlChildrenNode,
						 1)))
	    data->textinput_title = c;

	  /* description */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "description")
		   && !data->textinput_description
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->textinput_description = c;

	  /* name */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "name")
		   && !data->textinput_name
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->textinput_name = c;

	  /* link */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "link")
		   && !data->textinput_link
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->textinput_link = c;
	}
    }
}

static void
__mrss_parser_rss_skipHours (xmlDocPtr doc, xmlNodePtr cur, mrss_t * data)
{
  char *c;

  for (cur = cur->children; cur; cur = cur->next)
    {
      if (cur->type == XML_ELEMENT_NODE)
	{
	  if (!xmlStrcmp (cur->name, (xmlChar *) "hour")
	      && (c =
		  (char *) xmlNodeListGetString (doc, cur->xmlChildrenNode,
						 1)))
	    {
	      mrss_hour_t *hour;

	      if (!(hour = (mrss_hour_t *) malloc (sizeof (mrss_hour_t))))
		{
		  free (c);
		  return;
		}

	      memset (hour, 0, sizeof (mrss_hour_t));
	      hour->element = MRSS_ELEMENT_SKIPHOURS;
	      hour->hour = c;

	      if (!data->skipHours)
		data->skipHours = hour;
	      else
		{
		  mrss_hour_t *tmp;

		  tmp = data->skipHours;

		  while (tmp->next)
		    tmp = tmp->next;
		  tmp->next = hour;
		}
	    }
	}
    }
}

static void
__mrss_parser_rss_skipDays (xmlDocPtr doc, xmlNodePtr cur, mrss_t * data)
{
  char *c;

  for (cur = cur->children; cur; cur = cur->next)
    {
      if (cur->type == XML_ELEMENT_NODE)
	{
	  if (!xmlStrcmp (cur->name, (xmlChar *) "day")
	      && (c =
		  (char *) xmlNodeListGetString (doc, cur->xmlChildrenNode,
						 1)))
	    {
	      mrss_day_t *day;

	      if (!(day = (mrss_day_t *) malloc (sizeof (mrss_day_t))))
		{
		  free (c);
		  return;
		}

	      memset (day, 0, sizeof (mrss_day_t));
	      day->element = MRSS_ELEMENT_SKIPDAYS;
	      day->day = c;

	      if (!data->skipDays)
		data->skipDays = day;
	      else
		{
		  mrss_day_t *tmp;

		  tmp = data->skipDays;

		  while (tmp->next)
		    tmp = tmp->next;
		  tmp->next = day;
		}
	    }
	}
    }
}

static void
__mrss_parser_rss_item (xmlDocPtr doc, xmlNodePtr cur, mrss_t * data)
{
  char *c;
  char *attr;
  mrss_item_t *item;

  if (!(item = (mrss_item_t *) malloc (sizeof (mrss_item_t))))
    return;

  memset (item, 0, sizeof (mrss_item_t));
  item->element = MRSS_ELEMENT_ITEM;

  for (cur = cur->children; cur; cur = cur->next)
    {
      if (cur->type == XML_ELEMENT_NODE)
	{
	  /* title */
	  if (!xmlStrcmp (cur->name, (xmlChar *) "title")
	      && !item->title
	      && (c =
		  (char *) xmlNodeListGetString (doc, cur->xmlChildrenNode,
						 1)))
	    item->title = c;

	  /* link */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "link")
		   && !item->link
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    item->link = c;

	  /* description */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "description")
		   && !item->description
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    item->description = c;

	  /* source */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "source")
		   && !item->source
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    {
	      item->source = c;

	      if ((attr = (char *) xmlGetProp (cur, (xmlChar *) "url")))
		item->source_url = attr;
	    }

	  /* enclosure */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "enclosure")
		   && !item->enclosure
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    {
	      item->enclosure = c;

	      if ((attr = (char *) xmlGetProp (cur, (xmlChar *) "url")))
		item->enclosure_url = attr;

	      if ((attr = (char *) xmlGetProp (cur, (xmlChar *) "length")))
		item->enclosure_length = atoi (attr);

	      if ((attr = (char *) xmlGetProp (cur, (xmlChar *) "type")))
		item->enclosure_type = attr;
	    }

	  /* category */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "category")
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    {
	      mrss_category_t *category;

	      if (!
		  (category =
		   (mrss_category_t *) malloc (sizeof (mrss_category_t))))
		return;

	      memset (category, 0, sizeof (mrss_category_t));

	      category->element = MRSS_ELEMENT_CATEGORY;
	      category->category = c;

	      if ((attr = (char *) xmlGetProp (cur, (xmlChar *) "domain")))
		category->domain = attr;

	      if (!item->category)
		item->category = category;
	      else
		{
		  mrss_category_t *tmp;

		  tmp = item->category;
		  while (tmp->next)
		    tmp = tmp->next;
		  tmp->next = category;
		}
	    }

	  /* author */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "author")
		   && !item->author
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    item->author = c;

	  /* comments */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "comments")
		   && !item->comments
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    item->comments = c;

	  /* guid */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "guid")
		   && !item->guid
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    {
	      item->guid = c;

	      if ((attr =
		   (char *) xmlGetProp (cur, (xmlChar *) "isPermaLink"))
		  && !strcmp (attr, "false"))
		item->guid_isPermaLink = 0;
	      else
		item->guid_isPermaLink = 1;

	    }

	  /* pubDate */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "pubDate")
		   && !item->pubDate
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    item->pubDate = c;

	}
    }


  if (!data->item)
    data->item = item;
  else
    {
      mrss_item_t *tmp;

      tmp = data->item;

      while (tmp->next)
	tmp = tmp->next;
      tmp->next = item;
    }
}

static mrss_error_t
__mrss_parser_rss (mrss_version_t v, xmlDocPtr doc, xmlNodePtr cur,
		   mrss_t ** ret)
{
  mrss_t *data;
  char *c, *attr;

  if (!(data = (mrss_t *) malloc (sizeof (mrss_t))))
    return MRSS_ERR_POSIX;

  memset (data, 0, sizeof (mrss_t));
  data->element = MRSS_ELEMENT_CHANNEL;
  data->version = v;

  while (cur && xmlStrcmp (cur->name, (xmlChar *) "channel"))
    cur = cur->next;

  if (!cur)
    {
      free (data);
      return MRSS_ERR_PARSER;
    }

  for (cur = cur->children; cur; cur = cur->next)
    {
      if (cur->type == XML_ELEMENT_NODE)
	{
	  /* title */
	  if (!xmlStrcmp (cur->name, (xmlChar *) "title") && !data->title &&
	      (c =
	       (char *) xmlNodeListGetString (doc, cur->xmlChildrenNode, 1)))
	    data->title = c;

	  /* description */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "description")
		   && !data->description
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->description = c;

	  /* link */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "link") && !data->link
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->link = c;

	  /* language */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "language")
		   && !data->language
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->language = c;

	  /* rating */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "rating")
		   && !data->rating
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->rating = c;

	  /* copyright */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "copyright")
		   && !data->copyright
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->copyright = c;

	  /* pubDate */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "pubDate")
		   && !data->pubDate
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->pubDate = c;

	  /* lastBuildDate */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "lastBuildDate")
		   && !data->lastBuildDate
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->lastBuildDate = c;

	  /* docs */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "docs") && !data->docs
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->docs = c;

	  /* managingeditor */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "managingeditor")
		   && !data->managingeditor
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->managingeditor = c;

	  /* webMaster */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "webMaster")
		   && !data->webMaster
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->webMaster = c;

	  /* image */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "image"))
	    __mrss_parser_rss_image (doc, cur, data);

	  /* textinput */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "textinput"))
	    __mrss_parser_rss_textinput (doc, cur, data);

	  /* skipHours */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "skipHours"))
	    __mrss_parser_rss_skipHours (doc, cur, data);

	  /* skipDays */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "skipDays"))
	    __mrss_parser_rss_skipDays (doc, cur, data);

	  /* item */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "item"))
	    __mrss_parser_rss_item (doc, cur, data);

	  /* category */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "category")
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    {
	      mrss_category_t *category;

	      if (!
		  (category =
		   (mrss_category_t *) malloc (sizeof (mrss_category_t))))
		{
		  mrss_free ((mrss_generic_t *) data);
		  return MRSS_ERR_POSIX;
		}

	      memset (category, 0, sizeof (mrss_category_t));

	      category->element = MRSS_ELEMENT_CATEGORY;
	      category->category = c;

	      if ((attr = (char *) xmlGetProp (cur, (xmlChar *) "domain")))
		category->domain = attr;

	      if (!data->category)
		data->category = category;
	      else
		{
		  mrss_category_t *tmp;

		  tmp = data->category;
		  while (tmp->next)
		    tmp = tmp->next;
		  tmp->next = category;
		}
	    }

	  /* enclosure */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "cloud")
		   && !data->cloud
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    {
	      data->cloud = c;

	      if ((attr = (char *) xmlGetProp (cur, (xmlChar *) "domain"))
		  && !data->cloud_domain)
		data->cloud_domain = attr;

	      if ((attr = (char *) xmlGetProp (cur, (xmlChar *) "port"))
		  && !data->cloud_port)
		data->cloud_port = atoi (attr);

	      if ((attr =
		   (char *) xmlGetProp (cur, (xmlChar *) "registerProcedure"))
		  && !data->cloud_registerProcedure)
		data->cloud_registerProcedure = attr;

	      if ((attr = (char *) xmlGetProp (cur, (xmlChar *) "protocol"))
		  && !data->cloud_protocol)
		data->cloud_protocol = attr;
	    }

	  /* generator */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "generator")
		   && !data->generator
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->generator = c;

	  /* ttl */
	  else if (!xmlStrcmp (cur->name, (xmlChar *) "ttl")
		   && !data->ttl
		   && (c =
		       (char *) xmlNodeListGetString (doc,
						      cur->xmlChildrenNode,
						      1)))
	    data->ttl = atoi (c);

	}
    }

  *ret = data;

  return MRSS_OK;
}

static mrss_error_t
__mrss_parser (xmlDocPtr doc, mrss_t ** ret)
{
  mrss_error_t r = MRSS_ERR_VERSION;
  xmlNodePtr cur;
  char *c;

  if (!(cur = xmlDocGetRootElement (doc)))
    {
      xmlFreeDoc (doc);
      return MRSS_ERR_PARSER;
    }

  if (!xmlStrcmp (cur->name, (xmlChar *) "rss"))
    {
      if ((c = (char *) xmlGetProp (cur, (xmlChar *) "version")))
	{
	  /* 0.91 VERSION */
	  if (!strcmp (c, "0.91"))
	    r =
	      __mrss_parser_rss (MRSS_VERSION_0_91, doc, cur->children, ret);

	  /* 0.92 VERSION */
	  else if (!strcmp (c, "0.92"))
	    r =
	      __mrss_parser_rss (MRSS_VERSION_0_92, doc, cur->children, ret);

	  /* 2.0 VERSION */
	  else if (!strcmp (c, "2.0"))
	    r = __mrss_parser_rss (MRSS_VERSION_2_0, doc, cur->children, ret);

	  else
	    r = MRSS_ERR_VERSION;
	}

      else
	r = MRSS_ERR_VERSION;
    }

  else
    r = MRSS_ERR_PARSER;

  xmlFreeDoc (doc);
  return r;
}

/*************************** EXTERNAL FUNCTION ******************************/

mrss_error_t
mrss_parse_url (char *url, mrss_t ** ret)
{
  __mrss_download_t *download;
  xmlDocPtr doc;
  mrss_error_t err;

  if (!url || !ret)
    return MRSS_ERR_DATA;

  if (!(download = __mrss_download_file (url)))
    return MRSS_ERR_POSIX;

  if (!(doc = xmlParseMemory (download->mm, download->size)))
    {
      free (download->mm);
      free (download);
      return MRSS_ERR_PARSER;
    }

  if (!(err = __mrss_parser (doc, ret)))
    {
      if (!((*ret)->file = strdup (url)))
	{
	  *ret = NULL;
	  return MRSS_ERR_POSIX;
	}

      (*ret)->size = download->size;
    }

  free (download->mm);
  free (download);

  return err;
}

mrss_error_t
mrss_parse_file (char *file, mrss_t ** ret)
{
  xmlDocPtr doc;
  mrss_error_t err;
  struct stat st;

  if (!file || !ret)
    return MRSS_ERR_DATA;

  if (lstat (file, &st))
    return MRSS_ERR_POSIX;

  if (!(doc = xmlParseFile (file)))
    return MRSS_ERR_PARSER;

  if (!(err = __mrss_parser (doc, ret)))
    {
      if (!((*ret)->file = strdup (file)))
	{
	  *ret = NULL;
	  return MRSS_ERR_POSIX;
	}

      (*ret)->size = st.st_size;
    }

  return err;
}

mrss_error_t
mrss_parse_buffer (char *buffer, size_t size, mrss_t ** ret)
{
  xmlDocPtr doc;
  mrss_error_t err;

  if (!buffer || !size || !ret)
    return MRSS_ERR_DATA;

  if (!(doc = xmlParseMemory (buffer, size)))
    return MRSS_ERR_PARSER;

  if (!(err = __mrss_parser (doc, ret)))
    (*ret)->size = size;

  return err;
}

/* EOF */
