/* mRss - Copyright (C) 2005 bakunin - Andrea Marchesini 
 *                                <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "mrss.h"
#include "mrss_internal.h"

static void __mrss_write_string(void (*func)(void *, char *,...), void *obj, char *str) {
  int i;
  int len;
  char buf[1024];
  int j;

#define __NXML_CHECK_BUF \
           if(j==sizeof(buf)-1) { buf[j]=0; func(obj, "%s",buf); j=0; }

  if (!str)
    return;

  len = strlen (str);

  for (j = i = 0; i < len; i++)
    {
      if (str[i] == '\r')
        continue;

      else if (str[i] == '<')
        {
          buf[j++] = '&';
          __NXML_CHECK_BUF;
          buf[j++] = 'l';
          __NXML_CHECK_BUF;
          buf[j++] = 't';
          __NXML_CHECK_BUF;
          buf[j++] = ';';
          __NXML_CHECK_BUF;
        }

      else if (str[i] == '>')
        {
          buf[j++] = '&';
          __NXML_CHECK_BUF;
          buf[j++] = 'g';
          __NXML_CHECK_BUF;
          buf[j++] = 't';
          __NXML_CHECK_BUF;
          buf[j++] = ';';
          __NXML_CHECK_BUF;
        }

      else if (str[i] == '&')
        {
          buf[j++] = '&';
          __NXML_CHECK_BUF;
          buf[j++] = 'a';
          __NXML_CHECK_BUF;
          buf[j++] = 'm';
          __NXML_CHECK_BUF;
          buf[j++] = 'p';
          __NXML_CHECK_BUF;
          buf[j++] = ';';
          __NXML_CHECK_BUF;
        }

      else if (str[i] == '\'')
        {
          buf[j++] = '&';
          __NXML_CHECK_BUF;
          buf[j++] = 'a';
          __NXML_CHECK_BUF;
          buf[j++] = 'p';
          __NXML_CHECK_BUF;
          buf[j++] = 'o';
          __NXML_CHECK_BUF;
          buf[j++] = 's';
          __NXML_CHECK_BUF;
          buf[j++] = ';';
          __NXML_CHECK_BUF;
        }

      else if (str[i] == '\'')
        {
          buf[j++] = '&';
          __NXML_CHECK_BUF;
          buf[j++] = 'q';
          __NXML_CHECK_BUF;
          buf[j++] = 'u';
          __NXML_CHECK_BUF;
          buf[j++] = 'o';
          __NXML_CHECK_BUF;
          buf[j++] = 't';
          __NXML_CHECK_BUF;
          buf[j++] = ';';
          __NXML_CHECK_BUF;
        }

      else
        {
          buf[j++] = str[i];
          __NXML_CHECK_BUF;
        }
    }

  if (j)
    {
      buf[j] = 0;
      func (obj, "%s", buf);
      j = 0;
    }
}

static mrss_error_t
__mrss_write_real (mrss_t * mrss, void (*func) (void *, char *, ...),
		   void *obj)
{
  mrss_item_t *item;
  mrss_category_t *category;

  func (obj, "<?xml version=\"1.0\" encoding=\"ISO-8859-1\" ?>\n"
	"<rss version=\"");

  switch (mrss->version)
    {
    case MRSS_VERSION_0_91:
      func (obj, "0.91");
      break;

    case MRSS_VERSION_0_92:
      func (obj, "0.92");
      break;

    case MRSS_VERSION_2_0:
      func (obj, "2.0");
      break;
    }

  func (obj, "\">\n");

  func (obj, "  <channel>\n");

  if (mrss->title)
    {
      func (obj, "    <title>");
      __mrss_write_string (func, obj, mrss->title);
      func (obj, "</title>\n");
    }

  if (mrss->description)
    {
      func (obj, "    <description>");
      __mrss_write_string (func, obj, mrss->description);
      func (obj, "</description>\n");
    }

  if (mrss->link)
    {
      func (obj, "    <link>");
      __mrss_write_string (func, obj, mrss->link);
      func (obj, "</link>\n");
    }

  if (mrss->language)
    {
      func (obj, "    <language>");
      __mrss_write_string (func, obj, mrss->language);
      func (obj, "</language>\n");
    }

  if (mrss->rating)
    {
      func (obj, "    <rating>");
      __mrss_write_string (func, obj, mrss->rating);
      func (obj, "</rating>\n");
    }

  if (mrss->copyright)
    {
      func (obj, "    <copyright>");
      __mrss_write_string (func, obj, mrss->copyright);
      func (obj, "</copyright>\n");
    }

  if (mrss->pubDate)
    {
      func (obj, "    <pubDate>");
      __mrss_write_string (func, obj, mrss->pubDate);
      func (obj, "</pubDate>\n");
    }

  if (mrss->lastBuildDate)
    {
      func (obj, "    <lastBuildDate>");
      __mrss_write_string (func, obj, mrss->lastBuildDate);
      func (obj, "</lastBuildDate>\n");
    }

  if (mrss->docs)
    {
      func (obj, "    <docs>");
      __mrss_write_string (func, obj, mrss->docs);
      func (obj, "</docs>\n");
    }

  if (mrss->managingeditor)
    {
      func (obj, "    <managingeditor>");
      __mrss_write_string (func, obj, mrss->managingeditor);
      func (obj, "</managingeditor>\n");
    }

  if (mrss->webMaster)
    {
      func (obj, "    <webMaster>");
      __mrss_write_string (func, obj, mrss->webMaster);
      func (obj, "</webMaster>\n");
    }

  if (mrss->version == MRSS_VERSION_2_0)
    {
      if (mrss->generator)
	{
	  func (obj, "    <generator>");
	  __mrss_write_string (func, obj, mrss->generator);
	  func (obj, "</generator>\n");
	}

      if (mrss->ttl)
	func (obj, "    <ttl>%d</ttl>\n", mrss->ttl);
    }

  if (mrss->image_title || mrss->image_url || mrss->image_link
      || mrss->image_width || mrss->image_height || mrss->description)
    {

      func (obj, "    <image>\n");

      if (mrss->image_title)
	{
	  func (obj, "      <title>");
	  __mrss_write_string (func, obj, mrss->image_title);
	  func (obj, "</title>\n");
	}

      if (mrss->image_url)
	{
	  func (obj, "      <url>");
	  __mrss_write_string (func, obj, mrss->image_url);
	  func (obj, "</url>\n");
	}

      if (mrss->image_link)
	{
	  func (obj, "      <link>");
	  __mrss_write_string (func, obj, mrss->image_link);
	  func (obj, "</link>\n");
	}

      if (mrss->image_width)
	func (obj, "      <width>%d</width>\n", mrss->image_width);

      if (mrss->image_height)
	func (obj, "      <height>%d</height>\n", mrss->image_height);

      if (mrss->image_description)
	{
	  func (obj, "      <description>");
	  __mrss_write_string (func, obj, mrss->image_description);
	  func (obj, "</description>\n");
	}

      func (obj, "    </image>\n");
    }

  if (mrss->textinput_title || mrss->textinput_description
      || mrss->textinput_name || mrss->textinput_link)
    {

      func (obj, "    <textinput>\n");

      if (mrss->textinput_title)
	{
	  func (obj, "      <title>");
	  __mrss_write_string (func, obj, mrss->textinput_title);
	  func (obj, "</title>\n");
	}

      if (mrss->textinput_description)
	{
	  func (obj, "      <description>");
	  __mrss_write_string (func, obj, mrss->textinput_description);
	  func (obj, "</description>\n");
	}

      if (mrss->textinput_name)
	{
	  func (obj, "      <name>");
	  __mrss_write_string (func, obj, mrss->textinput_name);
	  func (obj, "</name>\n");
	}

      if (mrss->textinput_link)
	{
	  func (obj, "      <link>");
	  __mrss_write_string (func, obj, mrss->textinput_link);
	  func (obj, "</link>\n");
	}

      func (obj, "    </textinput>\n");
    }

  if ((mrss->version == MRSS_VERSION_0_92
       || mrss->version == MRSS_VERSION_2_0) && mrss->cloud)
    {
      func (obj, "    <cloud");

      if (mrss->cloud_domain)
	{
	  func (obj, " domain=\"");
	  __mrss_write_string (func, obj, mrss->cloud_domain);
	  func (obj, "\"");
	}

      if (mrss->cloud_port)
	func (obj, " port=\"%d\"", mrss->cloud_port);

      if (mrss->cloud_path)
	{
	  func (obj, " path=\"");
	  __mrss_write_string (func, obj, mrss->cloud_path);
	  func (obj, "\"");
	}

      if (mrss->cloud_registerProcedure)
	{
	  func (obj, " registerProcedure=\"");
	  __mrss_write_string (func, obj, mrss->cloud_registerProcedure);
	  func (obj, "\"");
	}

      if (mrss->cloud_protocol)
	{
	  func (obj, " protocol=\"");
	  __mrss_write_string (func, obj, mrss->cloud_protocol);
	  func (obj, "\"");
	}

      func (obj, ">%s</cloud>\n", mrss->cloud);
    }

  if (mrss->skipHours)
    {
      mrss_hour_t *hour;

      func (obj, "    <skipHours>\n");

      hour = mrss->skipHours;
      while (hour)
	{
	  func (obj, "      <hour>");
	  __mrss_write_string (func, obj, hour->hour);
	  func (obj, "</hour>\n");

	  hour = hour->next;
	}

      func (obj, "    </skipHours>\n");
    }

  if (mrss->skipDays)
    {
      mrss_day_t *day;

      func (obj, "    <skipDays>\n");

      day = mrss->skipDays;
      while (day)
	{
	  func (obj, "      <day>");
	  __mrss_write_string (func, obj, day->day);
	  func (obj, "</day>\n");
	  day = day->next;
	}

      func (obj, "    </skipDays>\n");
    }

  if ((mrss->version == MRSS_VERSION_0_92
       || mrss->version == MRSS_VERSION_2_0) && mrss->category)
    {
      category = mrss->category;
      while (category)
	{
	  func (obj, "    <category");

	  if (category->domain)
	    {
	      func (obj, " domain=\"");
	      __mrss_write_string (func, obj, category->domain);
	      func (obj, "\"");
	    }

	  func (obj, ">");
	  __mrss_write_string (func, obj, category->category);
	  func (obj, "</category>\n");

	  category = category->next;
	}
    }

  item = mrss->item;
  while (item)
    {
      func (obj, "    <item>\n");

      if (item->title)
	{
	  func (obj, "      <title>");
	  __mrss_write_string (func, obj, item->title);
	  func (obj, "</title>\n");
	}

      if (item->link)
	{
	  func (obj, "      <link>");
	  __mrss_write_string (func, obj, item->link);
	  func (obj, "</link>\n");
	}

      if (item->description)
	{
	  func (obj, "      <description>");
	  __mrss_write_string (func, obj, item->description);
	  func (obj, "</description>\n");
	}

      if (mrss->version == MRSS_VERSION_2_0)
	{
	  if (item->author)
	    {
	      func (obj, "      <author>");
	      __mrss_write_string (func, obj, item->author);
	      func (obj, "</author>\n");
	    }

	  if (item->comments)
	    {
	      func (obj, "      <comments>");
	      __mrss_write_string (func, obj, item->comments);
	      func (obj, "</comments>\n");
	    }

	  if (item->pubDate)
	    {
	      func (obj, "      <pubDate>");
	      __mrss_write_string (func, obj, item->pubDate);
	      func (obj, "</pubDate>\n");
	    }

	  if (item->guid)
	    {
	      func (obj, "      <guid isPermaLink=\"%s\">");
	      __mrss_write_string (func, obj, item->guid);
	      func (obj, "</guid>\n");
	    }

	}

      if (mrss->version == MRSS_VERSION_2_0
	  || mrss->version == MRSS_VERSION_0_92)
	{
	  if (item->source)
	    {
	      func (obj, "      <source");

	      if (item->source_url)
		{
		  func (obj, " url=\"");
		  __mrss_write_string (func, obj, item->source_url);
		  func (obj, "\"");
		}

	      func (obj, ">");
	      __mrss_write_string (func, obj, item->source);
	      func (obj, "</source>\n");
	    }

	  if (item->enclosure)
	    {
	      func (obj, "      <enclosure");

	      if (item->enclosure_url)
		{
		  func (obj, " url=\"");
		  __mrss_write_string (func, obj, item->enclosure_url);
		  func (obj, "\"");
		}

	      if (item->enclosure_length)
		func (obj, " length=\"%d\"", item->enclosure_length);

	      if (item->enclosure_type)
		{
		  func (obj, " type=\"");
		  __mrss_write_string (func, obj, item->enclosure_type);
		  func (obj, "\"");
		}

	      func (obj, ">");
	      __mrss_write_string (func, obj, item->enclosure);
	      func (obj, "</enclosure>\n");

	    }

	  category = item->category;
	  while (category)
	    {
	      func (obj, "      <category");

	      if (category->domain)
		{
		  func (obj, " domain=\"");
		  __mrss_write_string (func, obj, category->domain);
		  func (obj, "\"");
		}

	      func (obj, ">");
	      __mrss_write_string (func, obj, category->category);
	      func (obj, "</category>\n");

	      category = category->next;
	    }
	}

      func (obj, "    </item>\n");

      item = item->next;
    }

  func (obj, "  </channel>\n");
  func (obj, "</rss>\n");

  return MRSS_OK;
}

static void
__mrss_file_write (void *obj, char *str, ...)
{
  va_list va;

  va_start (va, str);
  vfprintf ((FILE *) obj, str, va);
  va_end (va);
}

static void
__mrss_buffer_write (void *obj, char *str, ...)
{
  va_list va;
  char s[4096];
  int len;
  char **buffer = (char **) obj;

  va_start (va, str);
  len = vsnprintf (s, sizeof (s), str, va);
  va_end (va);

  if (!*buffer)
    {
      if (!(*buffer = (char *) malloc (sizeof (char) * (len + 1))))
	return;

      strncpy (*buffer, s, len);
    }
  else
    {
      if (!(*buffer = (char *) realloc (*buffer,
					sizeof (char) * (strlen (*buffer) +
							 len + 1))))
	return;
      strncat (*buffer, s, len);
    }
}

/*************************** EXTERNAL FUNCTION ******************************/

mrss_error_t
mrss_write_file (mrss_t * mrss, char *file)
{
  FILE *fl;
  mrss_error_t ret;

  if (!mrss || !file)
    return MRSS_ERR_DATA;

  if (!(fl = fopen (file, "wb")))
    return MRSS_ERR_POSIX;

  ret = __mrss_write_real (mrss, __mrss_file_write, fl);
  fclose (fl);

  return ret;
}

mrss_error_t
mrss_write_buffer (mrss_t * mrss, char **buffer)
{
  if (!mrss || !buffer)
    return MRSS_ERR_DATA;

  return __mrss_write_real (mrss, __mrss_buffer_write, buffer);
}

/* EOF */
